@@ -1,22 +1,21 @@
 import sys
 import inspect
 from functools import partial
+from .cross import PY2
 
 
 __all__ = ['decorator', 'wraps', 'unwrap', 'ContextDecorator', 'contextmanager']
 
 
 def decorator(deco):
-    # Any arguments after first become decorator arguments
-    has_args = get_argcounts(deco) != (1, False, False)
-
-    if has_args:
-        # A decorator with arguments is essentialy a decorator fab
+    if has_single_arg(deco):
+        return wraps(deco)(make_decorator(deco))
+    else:
+        # Any arguments after first become decorator arguments
+        # And a decorator with arguments is essentialy a decorator fab
         def decorator_fab(*dargs, **dkwargs):
             return make_decorator(deco, dargs, dkwargs)
         return wraps(deco)(decorator_fab)
-    else:
-        return wraps(deco)(make_decorator(deco))
 
 
 def make_decorator(deco, dargs=(), dkwargs={}):
@@ -52,9 +51,18 @@ class Call(object):
             raise AttributeError(*e.args)
 
 
-def get_argcounts(func):
-    spec = inspect.getargspec(func)
-    return (len(spec.args), bool(spec.varargs), bool(spec.keywords))
+if PY2:
+    def has_single_arg(func):
+        spec = inspect.getargspec(func)
+        return len(spec.args) == 1 and not spec.varargs and not spec.keywords
+else:
+    def has_single_arg(func):
+        spec = inspect.signature(func)
+        if len(spec.parameters) != 1:
+            return False
+        arg = next(iter(spec.parameters.values()))
+        return arg.kind not in (arg.VAR_POSITIONAL, arg.VAR_KEYWORD)
+
 
 def get_argnames(func):
     func = getattr(func, '__original__', None) or unwrap(func)
@@ -65,20 +73,25 @@ def arggetter(func, _cache={}):
         return _cache[func]
 
     argnames = get_argnames(func)
-    argcount = len(argnames)
+    indexes = dict((name, i) for i, name in enumerate(argnames))
+    if func.__defaults__:
+        defaults = dict((argnames[-1-i], v) for i, v in enumerate(func.__defaults__))
+    else:
+        defaults = {}
 
     def get_arg(name, args, kwargs):
-        if name not in argnames:
+        if name not in indexes:
             raise TypeError("%s() doesn't have argument named %s" % (func.__name__, name))
         else:
-            if name in kwargs:
+            index = indexes[name]
+            if index < len(args):
+                return args[index]
+            elif name in kwargs:
                 return kwargs[name]
-            elif name in argnames:
-                index = argnames.index(name)
-                if index < len(args):
-                    return args[index]
-                else:
-                    return func.__defaults__[index - argcount]
+            elif name in defaults:
+                return defaults[name]
+            else:
+                raise TypeError("%s() missing required argument: '%s'" % (func.__name__, name))
 
     _cache[func] = get_arg
     return get_arg
